//================================================================
// nBase (nano Base) - Copyright (c) 1996   Daniele Giacomini
//
// Permission to use, copy, and distribute this software and its
// documentation for any purpose with or without fee is hereby
// granted, provided that the above copyright notice appear in all
// copies and that both that copyright notice and this permission
// notice appear in supporting documentation.
//
// Permission to modify the software is granted, but not the right to
// distribute the modified code. Modifications are to be distributed
// as patches to released version.
//
// THIS SOFTWARE MAY BE EXTREMLY DANGEROUS! FOR THIS SOFTWARE
// THERE IS ABSOLUTELY NO WARRANTY! THIS SOFTWARE IS PROVIDED
// "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR
// IMPLIED! IN NO EVENT WILL THE AUTHOR BE LIABLE TO YOU FOR
// ANY DAMAGES, INCLUDING LOST PROFITS, LOST SAVINGS, OR OTHER
// INCIDENTAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF THE USE
// OR INABILITY TO USE THIS SOFTWARE. IF THE LAW OF A COUNTRY
// OR A REGION DO NOT PERMIT TO SAVE THE AUTHOR FROM ALL POSSIBLE
// RESPONSABILITIES ON THIS SOFTWARE, IT MUST BE ASSUMED THAT
// THERE IS NO PERMISSION TO USE THIS SOFTWARE THERE.
//
//================================================================
// Send your comments, suggestions, bug reports and english text
// corrections to:
//
// Daniele Giacomini
// Via Turati, 15
// I-31100 Treviso
// Italy
//
//================================================================
// STANDARD INCLUDE
//================================================================
//#include "std.ch"
#include "common.ch"
#include "box.ch"
#include "color.ch"
#include "directry.ch"
#include "Fileio.ch"
#include "inkey.ch"
#include "memoedit.ch"
#include "rddsys.ch"
#include "setcurs.ch"
#include "set.ch"
#include "error.ch"
#include "dbstruct.ch"

//================================================================
// REQUEST 
//================================================================
// modify here if less or more RDDs (Replaceable Database Driver)
// are needed.
//request dbfcdx
//request dbfmdx
request dbfndx
request dbfntx // default
//request dbpx

//================================================================
// ASCII characters
//================================================================
#define NUL chr(00)  // NULl
#define SOH chr(01)  // Start Of Hiteration
#define STX chr(02)  // Start TeXt
#define ETX chr(03)  // End TeXt
#define EOT chr(04)  // End Of Transmission
#define ENQ chr(05)  // ENQuery
#define ACK chr(06)  // ACKnowledge
#define BEL chr(07)  // BELl
#define BS  chr(08)  // Back Space
#define HT  chr(09)  // Horizontal Tab
#define LF  chr(10)  // Line Feed
#define VT  chr(11)  // Vertical Tab
#define FF  chr(12)  // Form Feed
#define CR  chr(13)  // Carriage Return
#define SO  chr(14)  // Shift Out
#define SI  chr(15)  // Shif In
#define DLE chr(16)  // Data Loop Exit
#define DC1 chr(17)  // Dispositive Check 1
#define DC2 chr(18)  // Dispositive Check 2 
#define DC3 chr(19)  // Dispositive Check 3 
#define DC4 chr(20)  // Dispositive Check 4 
#define NAK chr(21)  // Not AKnowledge
#define SYN chr(22)  // SYNcronized
#define ETB chr(23)  // End Trasmission Block
#define CAN chr(24)  // CANcel
#define EM  chr(25)  // End Medium
#define SUB chr(26)  // SUBstitute
#define ESC chr(27)  // ESCape
#define FS  chr(28)  // File Separator
#define GS  chr(29)  // Group Separator
#define RS  chr(30)  // Record Separator
#define US  chr(31)  // Unit Separator

//================================================================
// ASCII STRINGS
//================================================================
#define NL( nTimes )    ( replicate ( CR + LF, nTimes ) )
#define ENDTEXT         NL(2) + "END OF TEXT"
#define SPACE_LEN       254 // max normal string, see memoline()
#define SPACE_STR       space(SPACE_LEN)

//================================================================
// COLORS 
//================================================================
#define COLOR_BW        "W/N, N/W, N, N, N/W"
#define COLOR_C         "GB+/N, N/GB+, N, N, N/GB+"
#define COLOR_DEFAULT   ( iif( iscolor(), COLOR_C, COLOR_BW ) )

//================================================================
// PROGRAM INFO/COPYRIGHT
//================================================================
#define PROGRAM_COPYRIGHT;   
    "nBase (nano Base) 1996.01.15 (c) 1996 Daniele Giacomini." +; 
    + NL(1) +;
    "nBase is freeware and comes with ABSOLUTELY NO WARRANTY!" +;
    + NL(1) +;
    "Daniele Giacomini" +; 
    + NL(1) +;
    "Via Turati, 15" +; 
    + NL(1) +;
    "I-31100 Treviso" +;
    + NL(1) +;
    "Italy"

//================================================================
// OTHERS
//================================================================
#define MAX_SELECT  250

#define DOT_MESSAGE_ERROR               "Error"
#define DOT_PROMPT_END_OF_PROGRAM       "Exit ?"
#define DOT_MENU_YES                    " YES "
#define DOT_MENU_NO                     " NO "
#define DOT_KEY_REMINDER;
    "[Esc] exit  [Pag]/[Pag] previous/next  [Enter] execute"

#define ERROR_MENU_CHOICE_BREAK         "Break"
#define ERROR_MENU_CHOICE_RETRY         "Retry"
#define ERROR_MENU_CHOICE_DEFAULT       "Default"
#define ERROR_MENU_CHOICE_QUIT          "Quit"

#define EX_EXIT_NORMAL          0
#define EX_EXIT_BREAK           1
#define EX_EXIT_NO_MACRO_FILE   2
#define EX_ERROR_END_ALONE      "END alone.;Missing IF/WHILE/..."
#define EX_ERROR_NO_CONDITION   "No condition!"
#define EX_ERROR_NO_END         "No END.;END not found while searching for it."
#define EX_ERROR_TOO_MUCH_END   "Too much END."
#define EX_ERROR_MACRO_FILE_NOT_FOUND   "Macro file not found!"
#define EX_MAX_NEST             256
#define EX_MENU_CHOICE_ABORT    "Abort"
#define EX_MENU_CHOICE_CONTINUE "Continue"
#define EX_PROMPT_MACRO_EXECUTION_ABORT "Abort or continue macro execution?"

#define RPT_DEFAULT_COMMAND_SYMBOL      "*"
#define RPT_DEFAULT_LPP                 60
#define RPT_ERROR_PRINT_COMMAND_UNCOMPLETED     "Print command uncompleted." 
#define RPT_ERROR_PRINT_FILE_NOT_FOUND          "File not found."
#define RPT_ERROR_PRINT_ISOLATED_END;    
    RPT_DEFAULT_COMMAND_SYMBOL + "END isolated." 
#define RPT_ERROR_PARENTESES    "Error with   delimiters."

#define TEXT_KEY_REMINDER;
   "[Esc] exit, []/[]/[Pag]/[Pag] move text."

#define TB_APP_MODE_ON( b )     ( b:cargo := TRUE  )
#define TB_APP_MODE_OFF( b )    ( b:cargo := FALSE )
#define TB_APP_MODE_ACTIVE( b ) ( b:cargo )
#define TB_ERROR_ARRAY_DIFFERENT_DIMENTIONS;    
                                "TB() - Different dimention arrays."
#define TB_ERROR_NO_COLUMNS     "TB() - No Columns."
#define TB_ERROR_NO_PICTURES    "TB() - No Pictures."
#define TB_ERROR_RECORD_LOCKED  "TB() - Record locked.;Try again later."
#define TB_MENU_YES             " YES "
#define TB_MENU_NO              " NO "
#define TB_PROMPT_DELETE_RECORD "Delete Record?"
#define TB_WINDOW_BOTTOM_MEMO   "[Esc] cancel  [Ctrl]+[W] save"

//================================================================
// STATIC GLOBAL
//================================================================
//...

//================================================================
// nBase - Main function
//================================================================
function nBase( cPar0,;
                cPar1,; 
                cPar2,; 
                cPar3,; 
                cPar4,; 
                cPar5,; 
                cPar6,; 
                cPar7,; 
                cPar8,; 
                cPar9 )
//
// nBase( [<cPar0>], [<cPar1>], [<cPar2>], [<cPar3>], 
//     [<cPar4>], [<cPar5>], [<cPar6>], [<cPar7>], 
//     [<cPar8>], [<cPar9>] ) --> NIL
//
// This is the main program function.
//
// <cPar0>      May contains the name of a macro prorgam (&)
//              to be executed. If no paramenter is given,
//              nBase() starts with the Dot() function.
//
// <cPar1> ... <cPar9>  Optional parameters to send to a macro
//                      program via public variables:
//                      c_Par1 ... c_Par9.
//

    local cOldScreen
    local cOldColor             := setcolor()
    local bOldErrorHandler      := errorblock( {|e| ErrorHandler(e)} )
    local nOldRow               := row()
    local nOldCol               := col()
    
    local Result

    // translate the calling parameters into public variables
    // for macro (&) use.
    public c_Par0
    public c_Par1
    public c_Par2
    public c_Par3
    public c_Par4
    public c_Par5
    public c_Par6
    public c_Par7
    public c_Par8
    public c_Par9

    Memvar->c_Par0 := cPar0
    Memvar->c_Par1 := cPar1
    Memvar->c_Par2 := cPar2
    Memvar->c_Par3 := cPar3
    Memvar->c_Par4 := cPar4
    Memvar->c_Par5 := cPar5
    Memvar->c_Par6 := cPar6
    Memvar->c_Par7 := cPar7
    Memvar->c_Par8 := cPar8
    Memvar->c_Par9 := cPar9

    begin sequence
        
        // Copyright
        QQOut( PROGRAM_COPYRIGHT )
        
        // Copyright notice is saved with savescreen()
        cOldScreen      := savescreen()

        // Default
        set( _SET_DATEFORMAT, "dd/mm/yyyy" )
        set( _SET_DEFAULT, "")
        set( _SET_DELETED, "OFF" )
        set( _SET_CONSOLE, "ON" )
        set( _SET_CONFIRM, "ON" )
        set( _SET_SCOREBOARD, "OFF" )
        set( _SET_INSERT, "ON" )
        setcolor( COLOR_DEFAULT )
        rddsetdefault( "dbfntx" )
   
        // Macro execution or Command line activation
        if !( empty( cPar0 ) )
            // The function EX() will check alone the file    
            // existance.
            EX( cPar0 )
        else
            // command line
            dot()
        end

        // end of program
        dbcloseall()
        dbcommitall()
        scroll()
        setpos(0,0)

    recover
        // nil
    end sequence
    errorblock(bOldErrorHandler)

    // restore old values
    setcolor( cOldColor )
    restscreen( NIL, NIL, NIL, NIL, cOldScreen )
    setpos( nOldRow, nOldCol )

    return NIL

//================================================================
// DOT COMMAND LINE
//================================================================
static function Dot()
//
// Dot () --> NIL
//
// Interactive command line. Only functions and calculations
// allowed.
//

   local getlist := {}
   local bOldErrorHandler
   local cOldScreen     := savescreen()

   local lMore     := .T.
   local cCommand  := ""
   local acCommand := {}
   local nMax      := 0
   local nInd      := 0

   // prepare command array and indexes
   aadd( acCommand, SPACE_STR )
   nMax := 1
   nInd := 1
      
   // clear screen
   scroll()
   
   // loop
   while .T.
      
      scroll( maxrow()-1,00,maxrow(),maxcol() )
      devpos( maxrow(), 00 )
      
      devout( ltrim(str(nInd)) + "/" + ltrim(str(nMax)) )
      devpos( maxrow(), 10 )
      devout( DOT_KEY_REMINDER )
      
      @maxrow()-1,00 get acCommand[nMax] picture "@s"+ltrim(str(maxcol()+1))+"@"
      
      readmodal(getlist)
      getlist := {}
      
      do case
      case lastkey() = K_PGUP  // previous
         if nInd > 1
            nInd--
            acCommand[nMax] := acCommand[nInd]
         end
      case lastkey() = K_PGDN  // next
         if nInd < nMax
            nInd++
            acCommand[nMax] := acCommand[nInd]
         end
      case lastkey() = K_ESC
         if alert( DOT_PROMPT_END_OF_PROGRAM,; 
                     { DOT_MENU_YES, DOT_MENU_NO } ) == 1
            exit
         end
      case lastkey() = K_ENTER
         cCommand := alltrim( acCommand[nMax] )
         if !( cCommand == "" )
            
            scroll(00,00,maxrow()-2,maxcol(), +2)
            devpos( maxrow()-3, 00 )
            devout( cCommand )
            
            bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
            begin sequence
               devpos( maxrow()-2, 00 )
               devout( &(cCommand) )
            recover
               devpos( maxrow()-2, 00 )
               devout( DOT_MESSAGE_ERROR )
            end
            errorblock(bOldErrorHandler)
            
            // array increment
            aadd( acCommand, SPACE_STR )
            nMax := len( acCommand )
            nInd := nMax
         end
      otherwise
         // do nothing
      end
   end

   restscreen( NIL, NIL, NIL, NIL, cOldScreen )    

   return NIL

//================================================================
// ERROR HANDLING
//================================================================
static function ErrorHandler(e)
//
// errorblock( {|e| ErrorHandler(e)} )
//

    local cMessage
    local aOptions
    local nChoice
   
    // zero division => 0
    if ( e:genCode == EG_ZERODIV )
        return (0)
    end

    // network open error
    if ( e:genCode == EG_OPEN .and. e:osCode == 32 .and. e:canDefault )
        neterr(.T.)
        return (.F.)
    end

    // lock error during APPEND BLANK
    if ( e:genCode == EG_APPENDLOCK .and. e:canDefault )
        neterr(.T.)
        return (.F.)
    end

    // build error message
    cMessage := ErrorMessage(e)

    // build options array
    aOptions := {ERROR_MENU_CHOICE_BREAK}
    if (e:canRetry)
        aadd(aOptions, ERROR_MENU_CHOICE_RETRY)
    end
    if (e:canDefault)
        aadd(aOptions, ERROR_MENU_CHOICE_DEFAULT)
    end
    aadd(aOptions, ERROR_MENU_CHOICE_QUIT)
   
    // alert box
    nChoice := 0
    while ( nChoice == 0 )
        nChoice := alert( cMessage, aOptions )
        if ( nChoice == NIL )
            exit
        end
    end

    // do as instructed
    do case
    case ( aOptions[nChoice] == ERROR_MENU_CHOICE_BREAK )
        break(e)
    case ( aOptions[nChoice] == ERROR_MENU_CHOICE_RETRY )
        return (.T.)
    case ( aOptions[nChoice] == ERROR_MENU_CHOICE_DEFAULT )
        return (.F.)
    otherwise
        // exit
    end

    cMessage += ErrorMoreInfo(e)

    Text( cMessage )

    // end of program
    errorlevel(1)
    quit

    return (.F.)

//----------------------------------------------------------------
static function ErrorMessage(e)

    local cMessage

    // start error message
    cMessage := if( e:severity > ES_WARNING, "Error ", "Warning " )

    // add subsystem name if available
    if ( ValType(e:subsystem) == "C" )
        cMessage += e:subsystem()
    else
        cMessage += "???"
    end

    // add subsystem's error code if available
    if ( ValType(e:subCode) == "N" )
        cMessage += ("/" + alltrim( str(e:subCode) ) )
    else
        cMessage += "/???"
    end

    // add error description if available
    if ( ValType(e:description) == "C" )
        cMessage += ("  " + e:description)
    end

    // add either filename or operation
    if ( !Empty(e:filename) )
        cMessage += (": " + e:filename)
    elseif ( !Empty(e:operation) )
        cMessage += (": " + e:operation)
    end

    // add dos error
    if !( Empty(e:osCode) )
        cMessage += ";(DOS Error " + alltrim( str(e:osCode) ) + ")"
    end

    return cMessage

//----------------------------------------------------------------
static function ErrorMoreInfo(e)

   local cMessage := ""
   local i := 0

   cMessage += NL(2) 
   
   // procedure names
   i := 2 // to jump ErrorHandler() and ErrorMoreInfo()
   while ( !Empty(procname(i)) )
      cMessage += rtrim( ProcName(i) ) +;
                  "(" + alltrim( str( ProcLine(i) ) ) + ")" +;
                  NL(1)
      i++
   end

   // actual alias
   if len( alias() ) == 0
      cMessage += NL(1)
      cMessage += "Actual area n. 0 - New Area"
   else
      cMessage += "Actual area n. " +; 
                  alltrim( str( select() ) ) +;
                  " - Alias: " + alias() + NL(1) +;
                  "----------"
   end
   
   // all alias info
   for i := 1 to MAX_SELECT
      if !( alias(i) == "" )
         cMessage += NL(1)
         cMessage += (i)->(DbStatus()) // alla fine c' un NL(1)        
         cMessage += "----------"
      end
   next

   return cMessage

//================================================================
// INFO
//================================================================
static function dbStatus(lRecursion)

   local cFileInfo := ""
   local cOrdKey := ""
   local cLinkExpr := ""
   local nOrder    := 0
   local nRelation := 0
   local cFilter   := ""
   local nRelSelect := 0

   // default
   if lRecursion == NIL
      lRecursion := .F.
   end

   if len( alias() ) == 0
      // No file is available now.
   else
      cFileInfo := "Select n. " + alltrim( str( select() ) ) + NL(1)
      cFileInfo += "Alias: " + alias() + NL(1)

      nOrder := 1
      while .T.
         cOrdKey := ordkey( nOrder )
         if cOrdKey == ""
            exit
         else
            cFileInfo += "Order n. " +;
                            alltrim( str( nOrder ) ) +;
                            ": " + ordname() + " " + cOrdKey + NL(1)
         end
         nOrder++
      end

      cFilter := dbfilter()
      if cFilter == ""
         // exit
      else
         cFileInfo += "Filter : " + cFilter + NL(1)
      end
      
      nRelation := 0
      while .T.
         cLinkExpr := dbrelation( ++nRelation )
         if cLinkExpr == ""
            exit
         else
            nRelSelect := dbrselect( nRelation )
            cFileInfo += "Relation n. " +; 
                            alltrim( str( nRelation ) ) +; 
                            " " + alias(nRelSelect) +;
                            ": " + cLinkExpr + NL(1) +;
                            "--------------------" + NL(1) +;
                            (nRelSelect)->(DbStatus(.T.)) + NL(1) +;
                            "--------------------" + NL(1) +;
                            "end Relation n. " +;
                            alltrim( str( nRelation ) ) +;
                            " " + alias(nRelSelect)
         end
      end

   end

   return cFileInfo

//****************************************************************
//****************************************************************
//================================================================
// TEXT DISPLAY
//================================================================
function Text( cText )
//
// Text( <cText> ) --> NIL
//
//
// <cText>      Text to display.  
//

    local nOldCursor := setcursor( SC_NONE )
    local cOldScreen := savescreen()
    local nOldRow    := row()
    local nOldCol    := col()

    begin sequence 
    
        if cText == NIL
            break
        end

        scroll()
        dispbox( maxrow()-1,00,maxrow()-1,maxcol() )
        devpos( maxrow(),00 ) 
        devout( TEXT_KEY_REMINDER )
        
        memoedit( cText + ENDTEXT, 00,00,maxrow()-2,maxcol(), .F. )

    end

    setcursor( nOldCursor )
    restscreen( NIL, NIL, NIL, NIL, cOldScreen )
    setpos( nOldRow, nOldCol )
   
    return NIL

//================================================================
// WAIT FUNCTIONS
// These functions must be "opened" and then "closed".
//================================================================
function WaitFileEval()
//
// WaitFileEval() --> .T.
//
// Shows a wait bar calculated with recno()/lastrec(),
//

   if recno() = lastrec()
      WaitProgress(2) // > 100%
   else
      WaitProgress( recno()/lastrec() )
   end

   return .T.

//================================================================
function WaitProgress( nPercent )
//
// WaitProgress ( [<nPercent>] ) --> NIL
//
// <nPercent>   the actual percent value (1 = 100%).
//              When nPercent > 1 the wait bar is closed.
//              When nPercent == 1 the wait bar is displayed
//              and closed.
//
// Shows a wait bar at the display bottom.
//
// This function must be "closed".
//
   
    static cOldScreen 

    local nElements
    local nBar
    local nOldRow    := row()
    local nOldCol    := col()

    begin sequence

        if cOldScreen == NIL
            cOldScreen := savescreen( maxrow(),00,maxrow(),maxcol() )
        end

        if nPercent == NIL .or.;
           nPercent > 1 
       
            restscreen( maxrow(),00,maxrow(),maxcol(), cOldScreen )
            cOldScreen := NIL

            break
        end
   
        nElements := maxcol()+1
        nBar := nElements*nPercent
        if nBar < 0
            nBar := 0
        end

        devpos( maxrow(),00 )
        devout( replicate( "", nBar ) )
        devout( replicate( "", nElements-nBar ) )

        // if 100%
        if nPercent >= 1
            restscreen( maxrow(),00,maxrow(),maxcol(), cOldScreen )
            cOldScreen := NIL
        end

    end

    // restore
    setpos( nOldRow, nOldCol )
    
    return NIL

//================================================================
function Wait( cMessage )
//
// Wait( [<cMessage>] ) --> .T.
//
// <cMessage>   Text message to show. If it is NIL, it closes
//              the message.
//
// It shows a wait message at the center of the screen.
//
// Wait() without argument, closes the message.
//
// This function must be closed.
//

    static cOldScreen
    static nTop
    static nLeft
    static nBottom
    static nRight
    
    local nOldCursor := setcursor( SC_NONE )
    local nOldRow    := row()
    local nOldCol    := col()

    local nWidth := 0

    local nLines
    local nLine

    begin sequence

        // wait close?
        if valtype( cMessage ) <> "C"
            if cOldScreen <> NIL
                restscreen( nTop,nLeft,nBottom,nRight, cOldScreen )
                cOldScreen := NIL
            end

            break
    
        end

        // already opened?
        if cOldScreen <> NIL
            restscreen( nTop,nLeft,nBottom,nRight, cOldScreen )
            cOldScreen := NIL
        end
    
        // line count
        cMessage := alltrim(cMessage)
        nLines := mlcount( cMessage, maxcol()-1 )
        nLine := 0
        while nLine < nLines
            nLine++
            nWidth := max( nWidth,; 
                len( rtrim( memoline(cMessage, maxcol()-1, nLine ) ) ); 
                         )
        end
   
        nTop        := int( ( (maxrow()+1) - (nLines+2) ) / 2 )
        nLeft       := int( ( (maxcol()+1) - (nWidth+2) ) / 2 ) 
        nBottom     := nTop + nLines+1
        nRight      := nLeft + nWidth+1

        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
    
        scroll( nTop, nLeft, nBottom, nRight )
        dispbox( nTop, nLeft, nBottom, nRight ) 
   
        nLine := 1
        while nLine <= nLines
            devpos( nTop+nLine, nLeft+1 )
            devout( memoline(cMessage, nWidth, nLine ) )
            nLine++
        end

    end sequence

    // restore
    setcursor( nOldCursor ) 
    setpos( nOldRow, nOldCol )
    
    return NIL

//================================================================
// PROMPT FUNCTIONS
//================================================================
function Accept( Field, cMessage )
//
// Accept(<Field>, [<cMessage>]) --> Field|NIL
//
// <Field>      the variable to be edited and returned
//
// <cMessage>   the message to show
//
// The function shows <cMessage> and waits for input to place
// into <Field>.
//

    local getlist := {}
    local cOldScreen    := ""
    local nOldCursor    := setcursor( SC_NORMAL ) 
    local nOldRow       := row()
    local nOldCol       := col()
    local nTop          := 0
    local nLeft         := 0
    local nBottom       := 0
    local nRight        := 0
    local nWidth        := 0
    local nLines
    local nLine
    local ReturnField   := NIL

    default cMessage to ""

    begin sequence

        if Field == NIL
            break
        end

        if valtype( cMessage ) <> "C"
            cMessage := NIL
        end
    
        // calculate field dimention
        do case 
        case valtype(Field) <> "C"  // data, numeric, logic
            nWidth := 40
        case valtype(Field) == "C" .and.; // character short
             len(Field) <= 40
            nWidth := 40
        case valtype(Field) == "C" .and.;  // character medium
             len(Field) > 40 .and.;
             len(Field) <= maxcol()-1
            nWidth := len(Field)
        otherwise
            nWidth := maxcol()-1
        end

        // message line count
        cMessage := alltrim(cMessage)
        nLines := mlcount( cMessage, maxcol()-1 )
        nLine := 0
        while nLine < nLines
            nLine++
            nWidth := max( nWidth,; 
                len( rtrim( memoline(cMessage, maxcol()-1, nLine ) ) ); 
                             )
        end
        
        // calculate window
        nBottom     := maxrow()
        nTop        := nBottom - ( nLines+1+1 )
        nLeft       := int( ( (maxcol()+1) - (nWidth+2) ) / 2 ) -0
        nRight      := nLeft + nWidth+1
        
        // save window area
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        
        // create window
        scroll( nTop, nLeft, nBottom, nRight )
        dispbox( nTop, nLeft, nBottom, nRight ) 
   
        nLine := 1
        while nLine <= nLines
            devpos( nTop+nLine, nLeft+1 )
            devout( memoline(cMessage, nWidth, nLine ) )
            nLine++
        end

        do case
        case valtype( Field ) == "C" 
            devpos( nTop+nLines+1, nLeft+1 )
            @row(),col() get Field picture "@s"+ltrim(str(nWidth))+"@"
        case valtype( Field ) == "N"
            devpos( nTop+nLines+1, nLeft+1 )
            @row(),col() get Field 
        case valtype( Field ) == "D"
            devpos( nTop+nLines+1, nLeft+1 )
            @row(),col() get Field 
        otherwise
            devpos( nTop+nLines+1, nLeft+1 )
            @row(),col() get Field 
        end
        
        readmodal( getlist )
        getlist := {}
        
        do case
        case lastkey() = K_ESC
            ReturnField := NIL
        otherwise
            ReturnField := Field
        end

    
    end sequence

    // restore
    setcursor( nOldCursor ) 
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )

    return ReturnField

//================================================================
// BROWSE 
//================================================================
function TB( nTop, nLeft, nBottom, nRight,;
                acCol, acColSayPic, acColHead,;
                alColCalc, nFreeze,;
                lModify, lAppend, lDelete,; 
                lAutoSort )
//
// TB( <nTop>, <nLeft>, <nBottom>, <nRight>,
//      <acCol>, <acColSayPic>, [<acColHead>], 
//      [<alColCalc>, [<nFreeze>],
//      [<lModify>], [<lAppend>], [<lDelete>],
//      [<lAutosort>] )  --> NIL
// 
// <nTop>, <nLeft>, <nBottom>, <nRight>         The display area 
//                                              used to browse.
// <acCol>              Column array.
// <acColSayPic>        Picture array.
// <acColHead>          Column head description array.
// <alColCalc>          Calculated column array: .T. menas
//                      calculated, .F. means editable.
// <nFreeze>            Number of columns to be left visible.
// <lModify>            Ability to modify the file.
// <lAppend>            Ability to append data.
// <lDelete>            Ability to delete records.
// <lAutosort>          Ability to refresh record position
//                      when the file is indexed.
//

    local getlist := {}
    local bOldErrorHandler
    local lReadExit
    local nOldCursor
    local nOldRow
    local nOldCol
    
    local nALen := 0

    local oBrowse
    local nKey
    local lMore := .T.
    local alColMemo
   
    default nTop     to 0
    default nLeft    to 0
    default nBottom  to maxrow()
    default nRight   to maxcol()
   
    default acColSayPic to; 
        iif( acCol == NIL, NIL, array( len( acCol ) ) )
    default acColHead to; 
        iif( acCol == NIL, NIL, aclone( acCol ) )
    default alColCalc to; 
        iif( acCol == NIL, NIL, afill( array( len( acCol ) ), .F. ) )
    default nFreeze   to 0
    default lModify   to .T.
    default lAppend   to .T.
    default lDelete   to .T.
    default lAutosort to .T.

    if acCol == NIL
        TBDefault( @acCol, @acColSayPic, @alColCalc, @acColHead )
        nFreeze := 1
    end
    if acCol == NIL
       alert( TB_ERROR_NO_COLUMNS )
       return NIL
    end
    
    alColMemo := array( len( acCol ) )

    if acColSayPic == NIL
        alert( TB_ERROR_NO_PICTURES ) 
        return NIL
    end   
    nALen := len( acCol )
    if !( nALen == len( acColSayPic ) .and.;
          nALen == len( acColHead ) .and.;
          nALen == len( alColCalc );
        )
        alert( TB_ERROR_ARRAY_DIFFERENT_DIMENTIONS )
        return NIL
    end

    // if it is not possible to modify...
    if !lModify
        lAppend := .F.
        lDelete := .F.
    end
   
    // if eof() it is better to move the record pointer
    if eof()
        dbgobottom()
    end
    if bof()
        dbgotop()
    end

    lReadExit           := readexit(.T.)          
    nOldCursor          := setcursor()
    nOldRow             := row()
    nOldCol             := col()
    
    bOldErrorHandler    := errorblock( {|e| ErrorHandler(e)} )
    begin sequence
   
        // Make a Tbrowse object for the current workarea
        oBrowse := TBNew( nTop, nLeft, nBottom, nRight,;
                        acCol, acColHead, alColCalc, alColMemo )
   
        TB_APP_MODE_OFF(oBrowse)
        
        // Use a custom 'skipper' to handle append mode (see below)
        oBrowse:skipBlock := { |nSkip| TBSkip( nSkip, oBrowse ) }
   
        // Define pictures
        TBColPicture(oBrowse, acColSayPic)
   
        // turn the cursor off while browsing
        setcursor(SC_NONE)
   
        // freeze columns
        oBrowse:freeze := nFreeze
   
        // Main loop
        while lMore
            // Don't let the cursor move into frozen columns
            if ( oBrowse:colPos <= oBrowse:freeze )
                oBrowse:colPos := ( oBrowse:freeze + 1 )
            end
            // Stabilize the display until it's stable or a key is pressed
            oBrowse:forceStable()
            // check hit top/bottom
            if ( oBrowse:hitBottom .and.;
                 !TB_APP_MODE_ACTIVE( oBrowse ) .and.;
                 lModify .and.; 
                 lAppend; 
               )
                TB_APP_MODE_ON( oBrowse )
                nKey := K_DOWN
            else
                // Make sure that the current record is showing
                // up-to-date data in case we are on a network.
                oBrowse:refreshCurrent():forceStable()
                // Everything's done -- just wait for a key
                nKey := inkey( 0 )
            end
      
            do case
            case ( nKey == K_ESC )
                // Esc means leave
                lMore := .F.
            otherwise // normally
                // Apply the key to the oBrowse
                TBApplyKey( oBrowse, nKey, acCol, alColCalc,; 
                    acColHead, alColMemo, lModify, lDelete,; 
                    lAutosort )
            end
      
        end

    recover
        // nil
    end sequence
    
    errorblock(bOldErrorHandler)
    readexit(lReadExit)          
    setcursor(nOldCursor)
    setpos( nOldRow, nOldCol )
    
    return NIL

//----------------------------------------------------------------
static function TBSkip(nSkip, oBrowse)
//
// TBSkip(nSkip, oBrowse) --> nSkippedRecords
//
// Handle record movement requests from the Tbrowse object.
//
   
    local lAppend       := TB_APP_MODE_ACTIVE( oBrowse )
    local nI            := 0
   
    do case
    case ( nSkip == 0 .or. lastrec() == 0 )
        // Skip 0 (significant on a network)
        dbSkip(0)
    case ( nSkip > 0 .and. !eof() )
        // Skip Forward
        while (nI < nSkip)     
            dbskip( 1 )
            if eof()
                iif(lAppend, nI++, dbskip(-1))
                exit
            end
            nI++
        end
    case ( nSkip < 0 )
        // Skip backward
        while ( nI > nSkip )           
            dbskip(-1)
            if bof()
               exit
            end
            nI--
        end
    end
   
   return nI

//----------------------------------------------------------------
static function TBApplyKey( oBrowse, nKey,; 
                    acCol, alColCalc, acColHead, alColMemo,; 
                    lModify, lDelete, lAutosort; 
                          )
//
// Apply one keystroke to the oBrowse.
//
   
    local bOldErrorHandler
    local lMemo := .F.
   
    // determina se si tratti di un campo memo o di
    // campo molto grande
    lMemo := alColMemo[oBrowse:colPos]
        
    do case
    case nKey == K_DOWN
        oBrowse:down()
    case nKey == K_PGDN
        oBrowse:pageDown()
    case nKey == K_CTRL_PGDN
        oBrowse:goBottom()
        TB_APP_MODE_OFF(oBrowse)
    case nKey == K_UP
        oBrowse:up()
        if TB_APP_MODE_ACTIVE( oBrowse )
            TB_APP_MODE_OFF( oBrowse )
            oBrowse:refreshAll()
        end
    case nKey == K_PGUP
        oBrowse:pageUp()
        if TB_APP_MODE_ACTIVE( oBrowse )
            TB_APP_MODE_OFF( oBrowse )
            oBrowse:refreshAll()
        end
    case nKey == K_CTRL_PGUP
        oBrowse:goTop()
        TB_APP_MODE_OFF( oBrowse )
    case nKey == K_RIGHT
        oBrowse:right()
    case nKey == K_LEFT
        oBrowse:left()
    case nKey == K_HOME
        oBrowse:home()
    case nKey == K_END
        oBrowse:end()
    case nKey == K_CTRL_LEFT
        oBrowse:panLeft()
    case nKey == K_CTRL_RIGHT
        oBrowse:panRight()
    case nKey == K_CTRL_HOME
        oBrowse:panHome()
    case nKey == K_CTRL_END
        oBrowse:panEnd()
    case nKey == K_RETURN
        do case
        case !lModify  // read only 
            // no edit allowed
            // (to view a memo, use a calculated column)
        case !alColCalc[oBrowse:colPos]
            if ( TBAlias(oBrowse, acCol, alColCalc) )->(rlock())
                // in caso di read only ed altre eventualit
                bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
                begin sequence
                    TBDoGet(oBrowse, acCol, lMemo, lAutosort)
                    dbcommit()
                recover
                    // nil
                end
                errorblock(bOldErrorHandler)
                dbunlock()
            else
                alert( TB_ERROR_RECORD_LOCKED )
            end
        otherwise
            // nil
        end
    case nKey == K_CTRL_Y .or.;
         nKey == K_CTRL_DEL     // delete
        do case
        case lModify .and.; 
             lDelete .and.; 
             !deleted()
            if ( TBAlias(oBrowse, acCol, alColCalc) )->(rlock()) .and.;
               alert( TB_PROMPT_DELETE_RECORD,; 
                   { TB_MENU_NO, TB_MENU_YES }; 
                    ) == 2
                dbdelete()
            else
                alert( TB_ERROR_RECORD_LOCKED )
            end
            dbcommit()
            dbunlock()
        case lModify .and.; 
             lDelete .and.; 
             deleted()
            if ( TBAlias(oBrowse, acCol, alColCalc) )->(rlock())
                recall
                dbcommit()
                dbunlock()
            else
                alert( TB_ERROR_RECORD_LOCKED )
            end
        otherwise
           // nil
        end
    otherwise
        do case
        case !lModify
            // no edit allowed
        case !alColCalc[oBrowse:colPos] .and. !lMemo
            // edit allowed
            keyboard chr(nKey) // repeated the key for the following TBDoGet
            if ( TBAlias(oBrowse, acCol, alColCalc) )->(rlock())
                bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
                begin sequence
                    TBDoGet(oBrowse, acCol, lMemo, lAutosort)
                    dbcommit()
                recover
                    // nil
                end
                errorblock(bOldErrorHandler)
                dbunlock()
            else
                alert( TB_ERROR_RECORD_LOCKED )
            end
        otherwise
            // do nothing
        end
    end
   
    return NIL

//----------------------------------------------------------------
static function TBDoGet(oBrowse, acCol, lMemo, lAutosort)
//
// Do a GET for the current column in the browse.
//
   
    local lFlag := .T.
    local oCol
    local getlist
    local nKey
    local nLen
    local lAppend
    local bSavIns
    local nSavRecNo := recno()
    local xNewKey
    local xSavKey
    local cOldScreen
    local cOldColor
    local nSetCursor
    local cMemoOld
    local cMemoNew


    // If we're at EOF we're adding the first record, so turn on append mode
    if eof()
        lAppend := TB_APP_MODE_ON(oBrowse)
    else
        lAppend := TB_APP_MODE_ACTIVE(oBrowse)
    end
   
    // Make sure screen is fully updated, dbf position is correct, etc.
    oBrowse:forceStable()
    if ( lAppend .and. (recno() == lastrec() + 1 ) )
        // dbAppend()
        dbappend()
        // if neterror, ignore
        neterr()
    end
   
    // Save the current record's key value (or NIL)
    xSavKey := iif(empty(indexkey()), NIL, &(indexkey()))
   
    // Get the current column object from the browse
    oCol := oBrowse:getColumn(oBrowse:colPos)
    
    if lMemo

        cOldColor     := setcolor()   
        cOldScreen    := savescreen( maxrow()/2, 00, maxrow(), maxcol() )
        nSetCursor    := setcursor ( SC_NORMAL )
    
        scroll( maxrow()/2, 00, maxrow(), maxcol() )
        colorselect( CLR_ENHANCED )
        devpos( maxrow()/2, 01)
        devout( padc( oCol:heading, maxcol()+1 ) )
        devpos( maxrow(), 01 )
        devout( padc( TB_WINDOW_BOTTOM_MEMO, maxcol()+1 ) )
        setcolor( cOldColor )
   
        cMemoOld := rtrim( &(acCol[oBrowse:colPos]) )
        cMemoNew := memoedit( cMemoOld,; 
                        maxrow()/2+1, 01, maxrow()-1, maxcol()-1,; 
                        .T., NIL, SPACE_LEN; 
                            )
        if ( cMemoOld <> cMemoNew )
            replace &(acCol[oBrowse:colPos]) with cMemoNew
            dbcommit()
        end
   
        // restore
        restscreen( maxrow()/2, 00, maxrow(), maxcol(), cOldScreen )
        setcursor ( nSetCursor )
    
    else
   
        // Get picture len to force scrolling if var is larger than window
        nLen := oBrowse:colWidth(oBrowse:colPos)
   
        // Create a corresponding GET
        getlist := { getnew( row(), col(),     ;
                       oCol:block,       ;
                       oCol:heading,     ;
                       oCol:picture,     ;
                       oBrowse:colorSpec ) }
   
        // Set insert key to toggle insert mode and cursor shape
        bSavIns := setkey(K_INS, { || InsToggle() })
   
        // Set initial cursor shape
        setcursor(iif(ReadInsert(), SC_INSERT, SC_NORMAL))
       
        readmodal(getlist)
        getlist := {}
    
        setcursor(SC_NONE)
        setkey(K_INS, bSavIns)

    end
    
    // Turn append mode off after each new record
    TB_APP_MODE_OFF(oBrowse)
   
    if lAutosort
        // Get the record's key value (or NIL) after the GET
        xNewKey := if(empty(indexkey()), NIL, &(indexkey()))
   
        oBrowse:inValidate()
        oBrowse:refreshAll():forceStable()
   
        // if the key has changed (or if this is a new record)
        if !(xNewKey == xSavKey) .or. (lAppend .and. xNewKey != NIL)
            // do a complete refresh
            oBrowse:refreshAll():forceStable()
            // Make sure we're still on the right record after stabilizing
            while &(indexkey()) > xNewKey .and. !oBrowse:hitTop()
                oBrowse:up():forceStable()
            end
        end
    end
   
    // Check exit key from get
    nKey := lastkey()
    if nKey == K_UP   .or.; 
       nKey == K_DOWN .or. ;
       nKey == K_PGUP .or.; 
       nKey == K_PGDN
        keyboard(chr(nKey))
    end
   
    return NIL

//----------------------------------------------------------------
static function InsToggle()   
//
// InsToggle()
//
// Toggle the global insert mode and the cursor shape.
//
//
    if readinsert()
        readinsert(FALSE)
        setcursor(SC_NORMAL)
    else
        readinsert(TRUE)
        setcursor(SC_INSERT)
    endif

    return NIL
//----------------------------------------------------------------
static function TBNew ( nTop, nLeft, nBottom, nRight,;
                        acCol, acColHead, alColCalc, alColMemo )

    local oBrowse
    local n
    local oColumn
    local cType
   
    // Start with a new browse object from TBrowseDB()
    oBrowse := TBrowseDB(nTop, nLeft, nBottom, nRight)

    // Add columns
    for n := 1 to len( acCol )
      
        // Make a new column 
        cType := valtype( &(acCol[n]) )  // speed up macro evaluation
        do case
        case alColCalc[n]  
            // Make a calculated field
            oColumn := TBAddCalCol( oBrowse, acColHead[n], acCol[n] )
            alColMemo[n] := .F.
        case ( cType == "M")  
            // Make a calculated field to activate the macro window
            oColumn := TBAddCalCol( oBrowse, acColHead[n],;
                           "padr(" + acCol[n] + ", 30 )" ) // shows 30 characters anyway
            alColMemo[n] := .T.
        case ( cType == "C") .and.;
             ( len( &(acCol[n]) ) > SPACE_LEN )
            // Make a calculated field to activate the macro window
            oColumn := TBAddCalCol( oBrowse, acColHead[n],; 
                          "left(" + acCol[n] + ", 30 )" ) // shows only 30 characters
            alColMemo[n] := .T.
        otherwise
            // Make a new normal field column
            oColumn := TBColumnNew( acColHead[n],;
                          TBFieldWBlock( acCol[n] ) ) 
            alColMemo[n] := .F.
        end
      
        // other attribute
        oColumn:headSep( "" )  // no head line separation
        oColumn:colSep( "" )
        oColumn:footSep( "" )  // no foot separation line
        oColumn:footing( "" )  // no foot
        
        oBrowse:addColumn( oColumn )


    next

    return oBrowse

//----------------------------------------------------------------
static function TBColPicture(oBrowse, acColSayPic)
//
// Set up pictures for the column.
//
   local n
   local oColumn
   default acColSayPic to NIL
   
   for n := 1 to oBrowse:colCount
      // Get (a reference to) the column
      oColumn := oBrowse:getColumn(n)
      // define picture
      oColumn:picture := acColSayPic[n]
   next
   
   return NIL

//----------------------------------------------------------------
static function TBFieldWBlock( cFieldName )
//   
// return the column code block
//
    local bFieldWBlock
    local cAlias
    local cField
    local nArrowStart
    local nSelect

    nArrowStart := at( "->", cFieldName )

    do case
    case nArrowStart == 0 // there is no arrow, so, no alias was specified.
        cAlias := NIL
        cField := alltrim( cFieldName )
    case nArrowStart > 0
        cAlias := substr( cFieldName, 1, nArrowStart-1 )
        cField := substr( cFieldName, nArrowStart+2 )
    end

    do case
    case cAlias == NIL
        bFieldWblock := fieldwblock( cField, select() )
    case upper(cAlias) == "FIELD"
        bFieldWblock := fieldwblock( cField, select() )
    case upper(cAlias) == "MEMVAR" .or. ;
        upper(cAlias) == "M"
        bFieldWblock := memvarblock( cField )
    otherwise
        nSelect := select(cAlias)
        bFieldWblock := fieldwblock( cField, nSelect )
    end
   
    return bFieldWBlock

//----------------------------------------------------------------
static function TBAlias( oBrowse, acCol, alColCalc )
//   
// return the alias name for the actual column
//
    local cAlias
    local nArrowStart

    nArrowStart := at( "->", acCol[oBrowse:colPos] )

    do case
    case alColCalc[oBrowse:colPos]
        cAlias := alias()
    case nArrowStart == 0 // there is no arrow, so, no alias was specified.
        cAlias := alias()
    case nArrowStart > 0
        cAlias := substr( acCol[oBrowse:colPos], 1, nArrowStart-1 )
    end

    return alltrim(cAlias)

//----------------------------------------------------------------
static function TBAddCalCol(oBrowse, cColHead, cCol)
//   
// create a calculated column
//
    local oColumn
   
    oColumn := TBColumnNew(cColHead, { || &(cCol) } ) // evaluate the macro each time
   
    return oColumn

//----------------------------------------------------------------
static function TBDefault( acCol, acColSayPic, alColCalc, acColHead )
//

    if ( alias() == "" )
         return NIL
    end
   
    // prepare arrays
    acCol       := {}
    acColSayPic := {}
    alColCalc   := {}
    acColHead   := {}

    // the first default column
    aadd ( acCol, "if( deleted(), '*', ' ') +" +;
                        " str(recno(), 5, 0)" )
    aadd( alColCalc, .T. )
    aadd( acColSayPic, "!!!!!!" )
    
    aadd( acColHead, "Rec. #" )
   
    // add the other columns
    TBDefColAdd( @acCol, @acColSayPic, @alColCalc, @acColHead )

    return NIL

//----------------------------------------------------------------
static function TBDefColAdd( acCol, acColSayPic, alColCalc,; 
                    acColHead )  
//
   
    local aStruct   := {}  
    local nFields   := 0  
    local nColIndex := 0   
    local nRelations := 0  
    local nIndRel := 0    
    local cChildAlias := ""

    default acCol        to {}
    default acColSayPic  to {} 
    default alColCalc    to {}
    default acColHead    to {}

    // determiniamo il numero di relazioni
    nRelations := 0
    while .T.
        nRelations++
        if dbrselect( nRelations ) > 0
            // prosegue
        else
            // deve decrementare nRelations perch l'ultimo non  vero
            nRelations--
            exit
        end
    end
    // adesso nRelations contiene il numero di relazioni con il file principale
   
    // si analizza il file di cui si dovr fare il browse
    aStruct := dbstruct()
    nFields := len( aStruct )

    // si compilano i vettori
    nColIndex := 1
    while nColIndex <= nFields
        aadd( acCol, Alias()+"->"+aStruct[nColIndex, 1] )
        aadd( alColCalc, .F. )
        aadd( acColSayPic, TBDefPicture( aStruct[nColIndex, 1],; 
                                         aStruct[nColIndex, 2],; 
                                         aStruct[nColIndex, 3],; 
                                         aStruct[nColIndex, 4] ) )
        aadd( acColHead, Alias()+"->"+aStruct[nColIndex, 1] )
        nColIndex++
    end
   
    // it adds eventual relations
    if nRelations > 0
        for nIndRel := 1 to nRelations
   
            // it determinates the Alias name for this relation
            cChildAlias := alias( dbrselect( nIndRel ) )

            // it adds the columns from the related Alias
            (cChildAlias)->(TBDefColAdd( @acCol, @acColSayPic,; 
                @alColCalc, @acColHead );
                           )
        next
    end

    return NIL

//----------------------------------------------------------------
static function TBDefPicture( cColumn, cColumnType, nColumnLen,; 
                    nColumnDec ) 

   local nLen := 0
   local cColSayPic := ""

   do case
   case cColumnType == "C"
      nLen := len( &(cColumn) )
      if nLen > 40
         cColSayPic := "@s40"
      else
         cColSayPic := replicate( "x", nLen )
      end
   case cColumnType == "N"
      if nColumnDec > 0
         // con decimale e virgola
         cColSayPic := ;
            replicate ( "9", nColumnLen-1-nColumnDec ) +;
            "." +;
            replicate ( "9", nColumnDec )
      else
         // intero
         cColSayPic := ;
            replicate ( "9", nColumnLen )
      end
   case cColumnType == "D"
      cColSayPic := "99/99/9999"
   case cColumnType == "L"
      cColSayPic := "L"
   case cColumnType == "M"
      cColSayPic := "@s40" // <Memo>
   end

   return cColSayPic
//================================================================
// DOC PRINT
//================================================================
functio rptMany( cText, bWhileCondition, bForCondition, cSymbol )
//
// rptMany( <cText>, [<bWhileCondition>], [<bForCondition>],
//          [<cSymbol>] ) --> NIL
//
// Prints with the nBase report standard the text contained into
// <cText> for every record in the active Alias matching the
// conditions.
//
// <cText>              the text to print.
// <bWhileCondition>    code block WHILE condition for record
//                      to be included.
// <bForCondition>      code block FOR condition for record
//                      to be included.
// <cSymbol>            the symbol for print commands contained
//                      into the text file.
//

    default bWhileCondition to {|| .T. }
    default bForCondition   to {|| .T. }
    default cSymbol         to RPT_DEFAULT_COMMAND_SYMBOL 

    dbgotop()
    while !eof() .and. eval(bWhileCondition)
        WaitFileEval()
        if eval(bForCondition)
            RPT( cText, NIL, NIL, NIL, SPACE_LEN, NIL, NIL, cSymbol )    
        end
        dbskip()
    end
    dbgotop()

    return NIL

//================================================================
function RPT( cText, nTotalLines, lMain, lHeadFoot,; 
              nLineLength, nTabSize, lWrap, cSymbol )
//
// RPT( <cText>, [<nTotalLines>], [<lMain>], [<lHeadFoot>], 
//      [<nLineLength>], [<nTabSize>], [<lWrap>], [<cSymbol>] )
//      --> NIL
//
// Prints with the nBase report standard the text contained into
// <cText> once.
//
// As this function is recursive, it is important to transfer the
// exact line dimention to the recursive call.
//
// <cText>              the text to be printed.
// <nTotalLines>        <cText> lines length, used with recursive
//                      call.
// <lMain>              true (.t.) means that it is the first
//                      call and not recursive. True is the
//                      default value.
// <lHeadFoot>          true (.t.) means that <cText> is a header 
//                      or footer.
// <nLineLength>        the maximal line length.
// <nTabSize>           tab length.
// <lWrap>              true (.t.) means that the text will be
//                      wrapped if necessary.
// <cSymbol>            the prefix simbol for report commands.
//

    local nLine := 0
    local cLine := ""
   
    local cSubText := ""
    local nSubTextLines := 0
    local cCommand := ""
    local cCondition := ""
    local nParameter := 0
    local cParameter := ""
    local xParameter
    local nNeeded  := 0

    local nOldRow    := row()
    local nOldCol    := col()

    static nLPP
    static nPrintedLine
    static cHeader
    static nHeadLines
    static cFooter
    static nFootLines
    static cLeft
    static lNewPage
    
    default nTotalLines  to mlcount( cText, nLineLength, nTabSize, lWrap )
    default lMain        to .T. // chiamata principale per l'intero documento
    default lHeadFoot    to .F. // se .T. indica che si sta stampando un head o foot
    default nLineLength  to SPACE_LEN
    default cSymbol      to RPT_DEFAULT_COMMAND_SYMBOL 
    
    begin sequence
    
        // if it is the first call
        if lMain == .T. .or. lMain == NIL
            // static variable initialization
            nLPP         := NIL
            nPrintedLine := NIL
            cHeader      := NIL
            nHeadLines   := NIL
            cFooter      := NIL
            nFootLines   := NIL
            cLeft        := NIL
            lNewPage     := NIL
            
            // public variable creation
            public n_Lines
        end
   
        // static defaults
        if nLPP == NIL
            nLPP := RPT_DEFAULT_LPP
        end
        if nPrintedLine == NIL
            nPrintedLine := 0
        end
        if cHeader == NIL
            cHeader := ""
        end
        if nHeadLines == NIL
            nHeadLines := 0
        end
        if cFooter == NIL
            cFooter := ""
        end
        if nFootLines == NIL
            nFootLines := 0
        end
        if cLeft == NIL
            cLeft := ""
        end
        if lNewPage   == NIL
            lNewPage := .T.
        end

        // if there are no more lines to print
        if nTotalLines <= 0
            break
        end
   
        // if first call, change row() col() position.
        if lMain
           setpos( maxrow(), maxcol()+1 )    // beiond the end 
        end
        
        // main loop
        for nLine = 1 to nTotalLines

            // line conuter correction.
            if nPrintedLine >= nLPP
                nPrintedLine := 0
            end

            // remaining lines calculation.
            Memvar->n_Lines := nLPP - nFootLines - nPrintedLines
      
            // next line
            cLine := memoline( cText, nLineLength, nLine, nTabSize, lWrap )    
      
            // line analisys
            do case
            case upper( alltrim(cLine) ) = cSymbol+"REM"
                // comment, the line is ignored
            case upper( alltrim(cLine) ) = cSymbol+"COMMENT"
                // comment, the line is ignored
            case upper( alltrim(cLine) ) = cSymbol+"END" 
                // if it appears here, it is an error, the line is ignored.
                alert( RPT_ERROR_PRINT_ISOLATED_END )
            case upper( alltrim(cLine) ) = cSymbol+"COMMAND"
                // Start a command line.
                // The command lines are isolated.
                cCommand := rptReadToEnd( cText, @nLine, NIL, nLineLength, nTabSize, lWrap, cSymbol ) 
                EXecute( cCommand, "REPORT COMMAND")
            case upper( alltrim(cLine) ) = cSymbol+"IF"
                cCondition := alltrim( substr( ltrim(cLine), 3+len(cSymbol) ) )
                cSubText := rptReadToEnd( cText, @nLine, @nSubTextLines, nLineLength, nTabSize, lWrap, cSymbol )       
                if cCondition == ""
                    // no condition
                    alert( alltrim(cLine) + NL(1) +;
                         RPT_ERROR_PRINT_COMMAND_UNCOMPLETED )
                    // line ignored
                else
                    if &(cCondition)
                        RPT( cSubText, nSubTextLines, .F., .F., nLineLength, nTabSize, lWrap, cSymbol )
                    end
                end
            case upper( alltrim(cLine) ) = cSymbol+"WHILE"
                cCondition := alltrim( substr( ltrim(cLine), 6+len(cSymbol) ) )
                cSubText := rptReadToEnd( cText, @nLine, @nSubTextLines, nLineLength, nTabSize, lWrap, cSymbol )       
                if cCondition == ""
                    // no condition
                else
                    while &(cCondition)
                        RPT( cSubText, nSubTextLines, .F., .F., nLineLength, nTabSize, lWrap, cSymbol )
                    end
                end
            case upper( alltrim(cLine) ) = cSymbol+"DBSKIP"
                // dbskip()
                // the eof() and bof() check MUST be done inside the document!!!
                cParameter := alltrim( substr( ltrim(cLine), 7+len(cSymbol) ) )
                if cParameter == ""
                    dbskip(+1)
                else
                    nParameter := &(cParameter)
                    dbskip( nParameter )
                end
            case upper( alltrim(cLine) ) = cSymbol+"LPP" 
                // lines per page
                cParameter := alltrim( substr( ltrim(cLine), 4+len(cSymbol) ) )
                if cParameter == ""
                    alert( alltrim(cLine) + NL(1) +;
                         RPT_ERROR_PRINT_COMMAND_UNCOMPLETED )
                    // line ignored
                else
                    nParameter := &(cParameter)
                    nLPP := nParameter
                end
            case upper( alltrim(cLine) ) = cSymbol+"PA"
                if (nPrintedLine > nHeadLines .and. nPrintedLine < nLPP)   //  in mezzo ad una pagina 
                    // allora deve stampare il footer
                    while nPrintedLine < (nLPP - nFootLines)
                        // stampa una riga vuota alla volta
                        // abilitando la stampa di head-foot
                        RPT( NL(1), 1, .F., .F., nLineLength, nTabSize, lWrap, cSymbol )
                    end
                    // il footer dovrebbe essere stampato automaticamente
                end
            case upper( alltrim(cLine) ) = cSymbol+"INSERT"
                cParameter := alltrim( substr( ltrim(cLine), 7+len(cSymbol) ) )
                if cParameter == ""
                    alert( alltrim(cLine) + NL(1) +;
                         RPT_ERROR_PRINT_COMMAND_UNCOMPLETED )
                    // la riga viene di fatto ignorata
                else
                    // se esiste il file completo di estensione viene incluso
                    if file( cParameter )
                        RPT( memoread(cParameter), NIL, .F., NIL, SPACE_LEN, NIL, NIL, cSymbol )
                    else
                        alert( alltrim(cLine) + NL(1) +;
                            RPT_ERROR_PRINT_FILE_NOT_FOUND ) 
                    end
                end
            case upper( alltrim(cLine) ) = cSymbol+"NEED"
                cParameter := alltrim( substr( ltrim(cLine), 5+len(cSymbol) ) )
                if cParameter == ""
                    alert( alltrim(cLine) + NL(1) +;
                        RPT_ERROR_PRINT_COMMAND_UNCOMPLETED ) 
                    // la riga viene di fatto ignorata
                else
                    nParameter := &(cParameter)
                    nNeeded := nParameter
                    if nPrintedLine + nNeeded > nLPP - nFootLines  // se uguale va ancora bene
                        // salta pagina in un modo un po' strano
                        RPT( cSymbol+"PA", 1, .F., .F., nLineLength, nTabSize, lWrap, cSymbol )
                    end
                end
            case upper( alltrim(cLine) ) = cSymbol+"HEAD"  
                cHeader := rptReadToEnd( cText, @nLine, @nHeadLines, nLineLength, nTabSize, lWrap, cSymbol ) 
            case upper( alltrim(cLine) ) = cSymbol+"FOOT"
                cFooter := rptReadToEnd( cText, @nLine, @nFootLines, nLineLength, nTabSize, lWrap, cSymbol ) 
            case upper( alltrim(cLine) ) = cSymbol+"LEFT"
                cParameter := alltrim( substr( ltrim(cLine), 5+len(cSymbol) ) )
                if cParameter == ""
                    alert( alltrim(cLine) + NL(1) +;
                        RPT_ERROR_PRINT_COMMAND_UNCOMPLETED )
                    // the line is ignored
                else   
                    xParameter := &(cParameter)
                    do case
                        case valtype(xParameter) == "N" .and.;
                            xParameter >= 0
                            cLeft := space(xParameter)
                        case valtype(xParameter) == "C"
                            cLeft := xParameter
                         otherwise
                        alert( alltrim(cLine) + NL(1) +;
                            RPT_ERROR_PRINT_COMMAND_UNCOMPLETED )
                        // the line is ignored
                    end
                end
            otherwise
                // Normal line to be printed.
                // Check if the head must be printed.
                if (nPrintedLine == 0) .and.;
                    !lHeadFoot
                    RPT( cHeader, nHeadLines, .F., .T., nLineLength, nTabSize, lWrap, cSymbol )
                end
                // It prints the line deleting extra characters on the right.
                // The extra space (the line is ever SPACE_LEN long) will 
                // produce extra white lines on the printed text.
                qqout( rtrim( cLeft + RPTLineTrans(cLine) ) )
                qout("")
                nPrintedLine++
                // It checks if the footer must be printed.
                if (nPrintedLine >= ( nLPP - nFootLines ) ) .and.;  
                   (lHeadFoot == .F.)
                 
                    RPT( cFooter, nFootLines, .F., .T., nLineLength, nTabSize, lWrap, cSymbol )
                    // After a footer there is ever a eject.
                    // The Eject command works only with printers
                    qqout( FF )  // form feed - chr(12)
                end
            end
   
        next

        if lMain .and.;    // main call
           (nPrintedLine > 0 .and. nPrintedLine < nLPP)   // in the middle of a page
            // Then the footer must be printed.
            while nPrintedLine < (nLPP - nFootLines)
                // It prints an empty line.
                RPT( NL(1), 1, .F., .F., nLineLength, nTabSize, lWrap, cSymbol )
            end
            // The footer should be automatically printed.
        end

    end sequence
    
    if lMain
        release c_Lines
        devpos( nOldRow, nOldCol )
    end
    
    return NIL

//----------------------------------------------------------------
static function rptReadToEnd( cText, nLine, nLines,; 
                              nLineLength, nTabSize,; 
                              lWrap, cSymbol )
//
// --> cBlock
//
// Returns a block of lines from <cText>.
// The block is delimited with:
//      *...Command
//              ...
//              ...
//      *end
//

    local cLine  := ""
    local cBlock := ""
    local nStartLine := nLine
    local nSubLines := 0
    local lStart    := .T.

    while .T.
        nLine++
        cLine := memoline( cText, nLineLength, nLine, nTabSize, lWrap  )
        do case
        case cLine == "" // end of block
            exit
        case upper( alltrim(cLine) ) = cSymbol+"END"
            exit
        case upper( alltrim(cLine) ) = cSymbol+"COMMAND" .or.;
             upper( alltrim(cLine) ) = cSymbol+"IF" .or.; 
             upper( alltrim(cLine) ) = cSymbol+"WHILE" .or.; 
             upper( alltrim(cLine) ) = cSymbol+"HEAD" .or.; 
             upper( alltrim(cLine) ) = cSymbol+"FOOT" 
            // The next *END must be jumped as it don't belongs
            // to this level.
            if lStart 
                // just started
                cBlock += rtrim(cLine)
            else
                cBlock += NL(1) + rtrim(cLine)
            end
            cBlock += NL(1) +;
                      rptReadToEnd( cText, @nLine, @nSubLines, nLineLength, nTabSize, lWrap, cSymbol ) +;
                      NL(1) + cSymbol+"END"
        otherwise
            if lStart 
                // Just started.
                cBlock += rtrim(cLine)
            else
                cBlock += NL(1) + rtrim(cLine)
            end
        end
        lStart := .F.
    end

    // At now, nLine is pinted to *END
   
    // nLines must count only the content not the commands: *... and *end.
    nLines := (nLine - nStartLine) -1

    // nLine and nLines are sent back
    return cBlock

//----------------------------------------------------------------
static function RPTLineTrans( cLine ) 
//
// --> cTranslatedLine
//
// Translates the <cLine> content replacing what is into
//  and , using Memvars and Fields.
//
// This function is recursive.
//

    local bSaveErrorHandler
    local bSav2ErrorHandler
    local nPosLeft   
    local nPosRight
    local cLineLeft
    local cLineCenter
    local cLineRight
    local cString
    local xContent

   
    nPosLeft := at( "", cLine )
    nPosRight := at( "", cLine )

    // if the line contains variables.
    do case
    case nPosLeft == 0 .and.;
         nPosRight == 0
        return cLine
    case nPosLeft == 0 .or.;    // se arriviamo qui non sono tutti e due uguali a zero
         nPosRight == 0
        alert( RPT_ERROR_PARENTESES )
        return cLine
    end

    // A variable string exists.
    cString := substr( cLine, nPosLeft+1 )
    cString := substr( cString, 1, at( "", cString ) - 1 )
    cString := alltrim( cString )

    // Left side determination.
    cLineLeft := substr( cLine, 1, nPosLeft-1 )

    // Variable transfer.
    bSaveErrorHandler := errorblock( {|| break(NIL)} )
    begin sequence
        xContent := &(cString) 
    recover
        // se il contenuto di cString non viene riconosciuto
        // genera un errore.
        bSaveErrorHandler := errorblock( {|| break(NIL)} )  // sta bene cos
        begin sequence
            // maybe, cString is a Memvar not jet defined.
            public &cString. := SPACE_STR
            &cString. := alltrim( Accept( &cString., "Insert " + cString ) )
            xContent := alltrim( &cString. )
        recover
            // But it wasn't a new variable.
            // It is written to help to find the mistake.
            xContent := "" + cString + ""
        end
        errorblock(bSav2ErrorHandler)

    end
    errorblock(bSaveErrorHandler)

    do case
    case valtype(xContent) == "C"
        cLineCenter := xContent     // senza alltrim
    case valtype(xContent) == "N"
        cLineCenter := alltrim( str( xContent ) )
    case valtype(xContent) == "D"
        cLineCenter := dtoc( xContent )
    case valtype(xContent) == "M"
        cLineCenter := xContent     // sentra trim
    end

    // Right extraction with recursion.
    cLineRight := RPTLineTrans( substr( cLine, nPosRight+1 ) )

    return cLineLeft + cLineCenter + cLineRight

//================================================================
// EX
//================================================================
function EX( cFileMacro )
//
// EX( <cFileMacro> ) --> nExitCode
//
// execute <cFileMacro>.
//
// <cFileMacro> The macro filename with extension.
//
//

    local cProcedure

    if file( cFileMacro )
        cProcedure := memoread( cFileMacro )
    else
        alert( cFileMacro + ";" +;
                EX_ERROR_MACRO_FILE_NOT_FOUND )
        return EX_EXIT_NO_MACRO_FILE
    end
   
    // start execution and terminate
    return EXecute( cProcedure, cFileMacro )

//----------------------------------------------------------------
static function EXecute( cCommands, cName ) 
//
// EXecute( <cCommands>, [<cName>] ) --> nExitCode
//
// <cCommands>  Variable containing many commands separated
//              by NL(1) ( = CR+LF).
//
// <cName>      A name to be used for error report duriing program
//              execution.
//

    local aNest         := array( EX_MAX_NEST )
    local nNest         := 0
    local nLine
    local cCommand      := ""
    local cCondition    := ""
    
    local nExitCode     := EX_EXIT_NORMAL

    begin sequence

        nLine := 1
        nNest := 1
        aNest[nNest] := {"MAIN", .T. }
        // [1]MAIN  [2]lContinue
      
        while .T.

            cCommand := EXCommandExtract( cCommands, @nLine ) 
            
            // line analisys
            
            if cCommand == NIL  // end of file
                nExitCode := EX_EXIT_NORMAL
                break
            end

            do case
            case upper( cCommand ) == "END"
                // close one nest level
                do case
                case aNest[nNest][1] == "MAIN"
                    // not allowed
                    if alert( cName + "(" + ltrim(str(nLine)) + ")" +";" +;
                           EX_ERROR_END_ALONE,; 
                               { EX_MENU_CHOICE_CONTINUE,; 
                                 EX_MENU_CHOICE_ABORT } ) == 2
                        nExitCode := EX_EXIT_BREAK
                        break
                    else
                        nLine++
                    end
                
                case aNest[nNest][1] == "IF" .or.;
                     aNest[nNest][1] == "THEN" .or.;
                     aNest[nNest][1] == "ELSE"
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    nLine++
                case aNest[nNest][1] == "BEGIN SEQUENCE" 
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    nLine++
                case aNest[nNest][1] == "WHILE"
                    if &(aNest[nNest][2])
                        // continue - go back
                        nLine := aNest[nNest][3]
                    else
                        // ok - end
                        aNest[nNest] := NIL
                        nNest--
                        nLine++
                    end
                
                case aNest[nNest][1] == "LOOP" 
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    // Previous level
                    // nLine remain on last END
                    do case
                    case aNest[nNest][1] == "WHILE"
                        if &(aNest[nNest][2])
                            // repeat loop
                            nLine := aNest[nNest][3]
                        else
                            // go to next END starting from next line
                            exJumpIt( cCommands, @nLine,; 
                                {|cLine| cLine=="END"} )
                            // stay, it will end naturally
                        end
                    otherwise
                        aNest[nNest][1] := "LOOP"
                        // go to next END starting from next line
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    end
                
                case aNest[nNest][1] == "BREAK" 
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    // nLine remain on last END
                    // previous level
                    do case
                    case aNest[nNest][1] == "BEGIN SEQUENCE"
                        // go to next END starting from next line
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    otherwise
                        aNest[nNest][1] := "BREAK"
                        // go to next END starting from next line
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    end
                case aNest[nNest][1] == "EXIT" 
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    // nLine remain on last END
                    // previous level
                    do case
                    case aNest[nNest][1] == "WHILE"  
                        // go to next END starting from next line
                        aNest[nNest][2] := ".F."
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    otherwise
                        aNest[nNest][1] := "BREAK"
                        // go to next END starting from next line
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    end
                case aNest[nNest][1] == "DO CASE" .or.;
                     aNest[nNest][1] == "ENDCASE"
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    nLine++
                end
            
            case upper( cCommand ) == "BEGIN SEQUENCE"  
                nNest++
                aNest[nNest] := {"BEGIN SEQUENCE"}
                // [1]BEGIN SEQUENCE
                nLine++
            case upper( cCommand ) == "BREAK"
                do case
                case aNest[nNest][1] == "BEGIN SEQUENCE"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                otherwise
                    aNest[nNest][1] := "BREAK"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "IF"  // no condition following
                // no condition
                alert( cName + "(" + ltrim(str(nLine)) + ")" +";" +;   
                    EX_ERROR_NO_CONDITION )
                exJumpIt( cCommands, @nLine,; 
                    {|cLine| cLine=="END"} )
                nLine ++
            case upper( left( cCommand, 3 ) ) == "IF "
                nNest++
                aNest[nNest] := {"IF" }
                // [1]IF|THEN|ELSE
                cCondition := alltrim( substr( ltrim(cCommand), 3 ) )
                if &(cCondition)
                    aNest[nNest][1] := "THEN"
                    nLine++
                else
                    aNest[nNest][1] := "ELSE"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="ELSE"}, {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "ELSE"
                if aNest[nNest][1] == "ELSE" // THEN wasn't executed
                    // ok execute ELSE
                    nLine++
                else
                    // THEN was executed and it must jumped
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "DO CASE"  
                nNest++
                aNest[nNest] := {"DO CASE"}
                // [1]DO CASE|ENDCASE
                nLine++
            case upper( cCommand ) == "CASE"  // no condition following
                // no condition
                alert( cName + "(" + ltrim(str(nLine)) + ")" +";" +;   
                    EX_ERROR_NO_CONDITION )
                exJumpIt( cCommands, @nLine,; 
                    {|cLine| left(cLine,5)=="CASE "},; 
                    {|cLine| cLine=="OTHERWISE"},; 
                    {|cLine| cLine=="END"} )
            case upper( left( cCommand, 5 ) ) == "CASE "
                do case
                case aNest[nNest][1] == "DO CASE" 
                    // ok try it
                    cCondition := alltrim( substr( ltrim(cCommand), 5 ) )
                    if &(cCondition)
                        aNest[nNest][1] := "ENDCASE"   // no more case
                        nLine++
                    else
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| left(cLine,5)=="CASE "},; 
                            {|cLine| cLine=="OTHERWISE"},; 
                            {|cLine| cLine=="END"} )
                    end
                otherwise // aNest[nNest][1] == "ENDCASE"
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "OTHERWISE"
                do case
                case aNest[nNest][1] == "DO CASE" 
                    // ok do it
                    nLine++
                otherwise // aNest[nNest][1] == "ENDCASE"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "WHILE" // no condition
                // no condition
                alert( cName + "(" + ltrim(str(nLine)) + ")" +";" +;   
                    EX_ERROR_NO_CONDITION )
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                nLine++
            case upper( left( cCommand, 6 ) ) == "WHILE "
                nNest++
                aNest[nNest] := {"WHILE", NIL, NIL }
                // [1]WHILE  [2]cCondition  [3]nLine1stStatement
                // save condition
                aNest[nNest][2] := alltrim( substr( ltrim(cCommand), 6 ) )
                // Save the first statement begin point (it is better).
                aNest[nNest][3] := nLine +1 
                // Check condition
                if &(aNest[nNest][2])
                    // ok, go on
                    nLine++
                else
                    aNest[nNest][2] := ".F."
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "LOOP"
                do case
                case aNest[nNest][1] == "WHILE"
                    if &(aNest[nNest][2])
                        // ok - loop
                        nLine := aNest[nNest][3] 
                    else   
                        // no - exit
                        aNest[nNest][2] := ".F."
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    end
                otherwise
                    aNest[nNest][1] := "LOOP"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "EXIT" 
                do case
                case aNest[nNest][1] == "WHILE"
                    aNest[nNest][2] := ".F."
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                otherwise
                    aNest[nNest][1] := "EXIT"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end

            otherwise
                // If we are here, it must be a function
                nExitCode := exOneCommand( cCommand, cName, nLine )
                if nExitCode == EX_EXIT_BREAK
                    break
                end
                nLine++
            end
        
        end
    
    end sequence

    return nExitCode
   
//----------------------------------------------------------------
static function exCommandExtract( cCommands, nLine )
//
// exCommandExtract( <cCommands>, @<nLine> ) --> NIL
//                                           --> cCommand
//
// It returns one complete command or NIL if end of file.
//
// <cCommands>  The variable containing the commands.
//
// <nLine>      The actual line position.
//
   
    local cCommand      := ""
    local nSemicolonPos := 0
   
    // reach next not-empty line.
    while .T.
        cCommand := memoline( cCommands, SPACE_LEN, nLine )
        do case
        case cCommand == ""
            // End of file reached
            return NIL 
        case !( empty( cCommand ) )
            // Cut eventual comment.
            cCommand := EXCommentCut( cCommand )
            // Cut unusefull blanks.
            cCommand := alltrim( cCommand )
            // Check if there is something.
            if !( cCommand == "" )
                // Not empty
                exit
            end
            // Otherwise repeat loop.
        end
        // Loop.
        nLine++
    end

    // At this point there is a command line
    
    // Semicolon
    if right( cCommand, 1 ) == ";"
        // it follows to the next line
        cCommand := left( cCommand, len( cCommand )-1 )
        nLine++
        // recursion
        cCommand += EXCommandExtract( cCommands, @nLine )
    end

    return cCommand

//----------------------------------------------------------------
static function EXCommentCut( cCommandLine )
//
// EXCommentCut( <cCommandLine> ) --> cCommandLine
//
// The function cuts the command line at the first
// occurence of "//" that is used as only comment
// indicator.
//
// The string "//" cannot be used for other purpose
// as the command line will be truncated there.
//
// <cCommandLine>       a line to be cutted.
//

    local nCommentPos

    nCommentPos := at( "//", cCommandLine )
    if nCommentPos > 0
        cCommandLine := substr( cCommandLine, 1, nCommentPos -1 )
    end
   
    return cCommandLine

//----------------------------------------------------------------
static function EXOneCommand( cCommand, cName, nLine )
//
// EXOneCommand( <cCommand>, <cName>, <nLine> ) --> nExitCode
//
// This function execute a command after cleaning.
//
// <cCommand>   The command to execute.
//
// <cName>      The procedure name to be used for error
//              documentation.
//
// <nLine>      Actual command line to be used for error
//              documentation.
//

    local bSaveErrorHandler
    local xResult
    local nExitCode := EX_EXIT_NORMAL

    // le righe vuote non devono essere eseguite
    if cCommand == ""
        return EX_EXIT_NORMAL
    end
                                          
    bSaveErrorHandler := errorblock( {|e| ErrorHandler( e, cName, nLine, cCommand )} )
    begin sequence
        xResult := &(cCommand)
    recover
        if alert( cName + " line: " + ltrim( str( nLine ) ) + ";" +;
                cCommand + ";;" +;
                EX_PROMPT_MACRO_EXECUTION_ABORT,; 
                    { EX_MENU_CHOICE_ABORT, EX_MENU_CHOICE_CONTINUE } ) == 1
            nExitCode := EX_EXIT_BREAK
        end
    end
    errorblock(bSaveErrorHandler)

    return nExitCode

//----------------------------------------------------------------
static function exJumpIt( cCommands, nLine,; 
                          bExit1, bExit2, bExit3 )
//
// --> NIL
//
// This function jumps to the statements whitch are not to be 
// executed according with the control structure used.
//

    local cLine
    local nLevel := 1
   
    default bExit1 to { | cLine | cLine == "END" }
    default bExit2 to bExit1
    default bExit3 to bExit2

    begin sequence

        while .T.
            nLine++
            cLine := EXCommandExtract( cCommands, @nLine )
            if cLine == NIL
                alert( EX_ERROR_NO_END )
                break
            end
            cLine := upper( cLine )
            
            if nLevel == 1 // start level
                if eval( bExit1, cLine ) .or.;
                   eval( bExit2, cLine ) .or.;
                   eval( bExit3, cLine )
                    break
                end
            end

            do case
            case cLine == "END"
                nLevel--
                if nLevel == 0
                    alert( EX_ERROR_TOO_MUCH_END )
                    break
                end
            case cLine            == "WHILE"            .or.;  // without condition
                 left( cLine, 6 ) == "WHILE "           .or.;  // with condition
                 cLine            == "IF"               .or.;  // without condition
                 left( cLine, 3 ) == "IF "              .or.;  // with condition
                 cLine            == "BEGIN SEQUENCE"   .or.;
                 cLine            == "DO CASE"

                nLevel++
            end
        end
    
    end sequence

    // nLine is pointed on END or beiond end of file
   
    return NIL

//****************************************************************
//****************************************************************
//================================================================
// COMMAND SUBSTITUTION FUNCTIONS
//================================================================
function bCompile( cBlockString )
//
// bCompile( <cBlockString> ) --> bBlock
//
// <bBlockString>       Block in string form.
//
// Compiles the string <cBlockString> and result a code block.
//
   
   if cBlockString <> NIL
      return &(cBlockString)
   else
      return NIL
   end

   return NIL

//================================================================
function CopyFile( cSource, cDestination )
//
// CopyFile( <cSource>, <cDestination> ) --> NIL
//
// Copy File command substitute.
//

    copy file (cSource) to (cDestination)

    return NIL

//================================================================
function dbApp( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest, cDriver )
//
// dbApp( <cFileName>, 
//              [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>], [<cDriver>] ) --> NIL
//
// Import records from a (.dbf) file.
//
// <cFileName>          the (.dbf) file name to use to 
//                      import data.
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
// <cDriver>            the RDD name to use.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbApp( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest, cDriver )

    return NIL

//================================================================
function dbContinue()
//
// dbContinue() --> NIL
//
// Continue command substitute.
//

    continue
    
    return NIL

//================================================================
function dbCopy( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest, cDriver )
//
// dbCopy( <cFileName>, 
//              [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>], [<cDriver>] ) --> NIL
//
// Export records to a new (.dbf) file.
//
// <cFileName>          the (.dbf) file name to create.
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
// <cDriver>            the RDD name to use.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbCopy( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest, cDriver )

    return NIL

//================================================================
function dbCopyStruct( cFileName, acFields )
//
// dbCopyStruct( <cFileName>, [<acFields>] )  --> NIL
//
// Copy the current (.dbf) structure to a new (.dbf) file.
//
    
    // change the following line if a new release uses 
    // a different internal function. 
    __dbCopyStruct( cFileName, acFields )
    
    return NIL

//================================================================
function dbCopyXStruct( cFileStructure )
//
// dbCopyXStruct( <cFileStructure> )  --> NIL
//
// Copy field definitions to a (.dbf) file.
//

    copy structure extended to (cFileStructure) 

    return NIL

//================================================================
function dbOldCreate( cFileName, cFileStructure,; 
                      cDriver, lNew, cAlias )
//
// dbOldCreate( <cFileName>, <cFileStructure>, 
//      [<cDriver>], [<lNew>], [<cAlias>] ) --> NIL
//
// Creates a new database file with structure defined by
// the specified extended file <cFileStructure>.
//
// The name "dbOldCreate" is so to make no confusion with the
// CA-Clipper function dbCreate that works with an Array for
// the data structue.
//
// <cFileName>          contains the file name to create.
// <cFileStructure>     contains the file containing the
//                      structure.
// <cDriver>            contains the driver name to be used.
// <lNew>               if true (.T.) it opend the new file
//                      into the next free work area, else
//                      it will be opened into the actura
//                      area.
// <cAlias>             it contains the alias to be used
//                      to indentify the new file.
//
    
    default lNew to .F.

    if lNew
        create (cFileName);
            from (cFileStructure);
            via cDriver;
            alias (cAlias);
            new
    else
        create (cFileName);
            from (cFileStructure);
            via cDriver;
            alias (cAlias)
    end

    return NIL

//================================================================
function dbDelim( lCopy, cFileName, cDelimiter, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )
//
// dbDelim( <lCopy>, <cFileName>, [<cDelimiter>], 
//              [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>] ) --> NIL
//
// Export records to a new ASCII file.
// Import records from a ASCII file.
//
// <lCopy>              if true (.T.) it exports, else it 
//                      imports data.
// <cFileName>          the ASCII file name to create or the 
//                      ASCII file name to use to import data.
// <cDelimiter>         the delimiter used.
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbDelim( lCopy, cFileName, cDelimiter, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )

    return NIL

//================================================================
function dbJoin( cAlias, cFileName, acFields, bForCondition )
//
// dbJoin( <cAlias>, <cFileName>, [<acFields>], 
//      [<bForCondition>] ) --> NIL
//
// Generates a new (.dbf) file by joining the active Alias
// with a second alias <cAlias> and eliminating records that
// do not meet the condition stated into the code block
// <bForCondition>.
//
// <cAlias>             the other Alias containing the data
//                      to use to join.
// <cFileName>          the (.dbf) file to create.
// <acFields>           fields to be involved.
// <bForCondition>      code block FOR.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbJoin( cAlias, cFileName, acFields, bForCondition )

    return NIL

//================================================================
function dbLabelForm( cLabel, lToPrinter, cFileName,;
             lNoConsole, bForCondition, bWhileCondition,; 
             nNextRecords, nRecord, lRest,; 
             lSample )
//
// dbLabelForm( <cLabel>, [<lToPrinter>], [<cFileName>],
//     [<lNoConsole>], [<bForCondition>], [<bWhileCondition>], 
//     [<nNextRecords>], [<nRecord>], [<lRest>], 
//     [<lSample> ) --> NIL
//
// Displays or prints lables form a definition held in a 
// label (.lbl) file for a range of records in the active 
// Alias.
//
// <cLabel>             the label (.lbl) file name.
// <lToPrinter>         if true (.T.) the labels will be
//                      printed.
// <cFileName>          the text file name to create 
//                      containing the lables.
// <lNoConsole>         if true (.T.) the labels will not
//                      displayed on the console.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
// <lSample>            if true (.T.), only test labels
//                      of asterisks are printed.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __LabelForm( cLabel, lToPrinter, cFileName,;
        lNoConsole, bForCondition, bWhileCondition,; 
        nNextRecords, nRecord, lRest,; 
        lSample )
    
    return NIL

//================================================================
function dbList( lToDisplay, abListColumns, lAll,;
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest,; 
                lToPrinter, cFileName )
//
// dbList( [<lToDisplay>], <abListColumns>, [<lAll>],
//     [<bForCondition>], [<bWhileCondition>], 
//     [<nNextRecords>], [<nRecord>], [<lRest>], 
//     [<lToPrinter>], [<cFileName>] ) --> NIL
//
// Displays or prints the result of one or more expressions
// listed into <abListColumns> for a range of records in the
// active Alias.
//
// <lToDisplay>         if true (.T.) it displays the list.
// <abListColumns>      every element contains the expression
//                      (in code block form) to be listed.
// <lAll>               if true (.T.) it lists all records.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
// <lToPrinter>         if true (.T.) it prints the list.
// <cFileName>          the text file to create containing
//                      the list.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbList( lToDisplay, abListColumns, lAll,;
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest,; 
                lToPrinter, cFileName )

    return NIL

//================================================================
function dbLocate( bForCondition, bWhileCondition,; 
    nNextRecords, nRecord, lRest; 
                 )
//
// dbLocate( [<bForCondition>], [<bWhileCondition>], 
//     [<nNextRecords>], [<nRecord>], [<lRest>] ) --> NIL
//
// Positions the record pointer to the first record
// in the active Alias that matches the specified condition
// within the given scope.
//
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
//
    
    default bForCondition to { || .T. }
    default bWhileCondition to { || .T. }
    
    // change the following line if a new release uses 
    // a different internal function. 
    __dbLocate( bForCondition, bWhileCondition,; 
        nNextRecords, nRecord, lRest; 
              )

    return NIL

//================================================================
function dbPack()
//
// dbPack() --> NIL
//
// Pack command substitute.
//

    pack
    
    return NIL

//================================================================
function dbReportForm( cForm, lToPrinter, cFileName,;
             lNoConsole, bForCondition, bWhileCondition,; 
             nNextRecords, nRecord, lRest,; 
             lPlain, cHeading,;
             lNoEject, lSummary )
//
// dbReportForm( <cForm>, [<lToPrinter>], [<cFileName>],
//     [<lNoConsole>], [<bForCondition>], [<bWhileCondition>], 
//     [<nNextRecords>], [<nRecord>], [<lRest>], 
//     [<lPlain>], [<cHeading>],
//     [<lNoEject>, [<lSummary>] ) --> NIL
//
// Displays or prints a tabular and optionally grouped 
// report with page and column headings for a range of
// records in the active Alias.
// The report definition is held in a report (.frm) file.
//
// <cForm>              the report (.frm) file name.
// <lToPrinter>         if true (.T.) the report will be
//                      printed.
// <cFileName>          the text file name to create 
//                      containing the report.
// <lNoConsole>         if true (.T.) the report will not
//                      displayed on the console.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
// <lPlain>             if true (.T.), the report will
//                      not have page breaks.
// <cHeading>           contains the header printed each
//                      page.
// <lNoEject>           if true (.T.) it suppresses the
//                      initial page eject.
// <lSummary>           if true (.T.), only group, subgroup, 
//                      and grand total lines are printed.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __ReportForm( cForm, lToPrinter, cFileName,;
        lNoConsole, bForCondition, bWhileCondition,; 
        nNextRecords, nRecord, lRest,; 
        lPlain, cHeading,;
        lNoEject, lSummary )
    
    return NIL

//================================================================
function dbSDF( lCopy, cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )
//
// dbSDF( <lCopy>, <cFileName>, 
//              [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>] ) --> NIL
//
// Export records to a new ASCII file.
// Import records from a ASCII file.
//
// <lCopy>              if true (.T.) it exports, else it 
//                      imports data.
// <cFileName>          the ASCII file name to create or the 
//                      ASCII file name to use to import data.
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbSDF( lCopy, cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )

    return NIL

//================================================================
function dbSort( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )
//
// dbSort( <cFileName>, [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>] ) --> NIL
//
// Copies records within the specified scope and condition
// from the current work area to another database file
// sorted according to the specified fields.
//
// <cFileName>          the (.dbf) file name to create. 
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbSort( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )

    return NIL

//================================================================
function dbTotal( cFileName, bKey, acFields,;
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )
//
// dbTotal( <cFileName>, <bKey>, [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>] ) --> NIL
//
// Summerizes records in the current work area by key value,
// summing the specified numeric fields and then copying
// summary records to a second database file.
//
// <cFileName>          the (.dbf) file name to create. 
// <bKey>               code block Key expression.
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbTotal( cFileName, bKey, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )

    return NIL

//================================================================
function dbUpdate( cAlias, bKey, lRandom, bReplacement )
//
// dbUpdate( <cAlias>, <bKey>, [<lRandom>], 
//              [<bReplacement>] ) --> NIL
//
// Updates the active Alias with the data contained into
// <cAlias> using the code block <bReplacement>.
//
// <cAlias>             the other Alias containing the data
//                      to use to update the active Alias.
// <bKey>               code block Key expression.
// <lRandom>            if true (.T.), then RANDOM.
// <bReplacement>       code block to be executed for the
//                      records with a matching key.
//
// Example:
//      dbUpdate( "INVOICE", {|| LAST}, .T.,;
//              {|| FIELD->TOTAL1 := INVOICE->SUM1,;
//                  FIELD->TOTAL2 := INVOICE->SUM2 } )
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbUpdate( cAlias, bKey, lRandom, bReplacement )

    return NIL

//================================================================
function dbZap()
//
// dbZap() --> NIL
//
// Zap command substitute.
//

    zap
    
    return NIL

//================================================================
function Get( nRow, nCol, cVar,; 
    cGetPicture, cColorString, bPreExpression, bValid; 
            )
//
// like @...get
//
   memvar getlist
   
   default cColorString   to setcolor()
   default bPreExpression to { || .T. }
   default bValid         to { || .T. }

   setpos( nRow, nCol ) 
   aadd( GetList, _GET_(; 
       &cVar., cVar, cGetPicture, bValid, bPreExpression; 
                       ):display(); 
       )
   atail(GetList):colorDisp(cColorString)
   
   return NIL

//================================================================
function Keyboard( cString )
//
// Keyboard( [<cString>] ) --> NIL
//
// Keyboard command substitute.
//
    default cString to ""

    keyboard cString

    return NIL

//================================================================
function MemPublic( cVarName )
//
// MemPublic( <cVarName> ) --> NIL
//
// <cVarName>   Contains the variable name.
//
// Creates the public variable named into <cVarName>.
//
   
   if cVarName <> NIL
      public &cVarName.
   else
      // nil
   end

   return NIL

//================================================================
function MemRelease( cMemvar )
//
// Release( <cMemvar> ) --> NIL
//
// Release command substitute.
//

    release &cMemvar.

    return NIL

//================================================================
function MemRestore( cMemFileName, lAdditive )
//
// MemRestore( <cMemFileName>, [<lAdditive>] ) --> NIL
//
// Restore From command substitute.
//

    default lAdditive to .F.

    // change the following line if a new release uses 
    // a different internal function. 
    __MRestore( cMemFileName, lAdditive )

    return NIL

//================================================================
function MemSave( cMemFileName, cSkeleton, lLike )
//
// MemSave( <cMemFileName>, [<cSkeleton>], [<lLike>] ) --> NIL
//
// Save To command substitute.
//
// <cMemFileName>       (.mem) file name to create.
// <cSkeleton>          memory names skeleton.
// <lLike>              true (.T.) means LIKE SKELETON,
//                      false (.F.) means EXCEPT SKELETON.
//

    default cSkeleton to "*"
    default lLike to .T.

    if lLike
        save to &cMemFileName. all like &cSkeleton.
    else
        save to &cMemFileName. all except &cSkeleton.
    end

    return NIL

//================================================================
function Quit()
//
// Quit() --> NIL
//
// Quit command substitute.
//

    quit
    
    return NIL

//================================================================
function Run( cCommand )
//
// Run( <cCommand> ) --> NIL
//
// <cCommand>   String to execute by OS.
//
// Run command substitute.
//

   if cCommand <> NIL
      run (cCommand)
   else   
      // nil
   end

   return NIL

//================================================================
function Say( nRow, nCol, cVar, cSayPicture, cColorString )
//
// like @...say
//
   memvar getlist
   
   default cColorString   to setcolor()

   
   devpos( nRow, nCol ) 
   if cSayPicture == NIL
      devout( cVar, cColorString )
   else
      devoutpict( cVar, cSayPicture, cColorString )
   end

   return NIL

//================================================================
function SetFunction( nFunctionKey, cString )
//
// SetFunction( <nFunctionKey>, [<cString>] ) --> NIL
//
// Set Function command substitute.
//
    default cString to ""

    set function nFunctionKey to cString

    return NIL

//================================================================
// END
//================================================================

